  check_required_tests:
    name: All {{ group }} tests passed
    needs: test-{{ group | replace(" ", "_") | lower }}
    # Always run this, even if a dependent job failed
    if: always()
    runs-on: ubuntu-22.04
    steps:
      - name: Check for failures
        if: needs.test-{{ lowercase_group }}.result != 'success'
        run: |
          echo "One of the dependent jobs has failed. You may need to re-run it." && exit 1
